
###########################################################################
# Copyright 2015 Christoph Huett  (email : christoph.huett@uni-koeln.de)
#
# This software is licensed under The MIT License (MIT) 
# http://opensource.org/licenses/MIT
#
# It works with ArcGIS pro X64 Python 3.4
#
############################################################################

import arcpy, numpy, os, shutil#, win32com.client,
from dbfpy3 import dbf
#Install this package using this link
#https://github.com/phargogh/dbfpy3


def createsubfolderifnotthere (base,name):
    #debug = 0
    path = os.path.join(base,name)
    if (not(os.path.exists (path))):
        os.makedirs(path)
        if debug: print(name, " created")


def classify (outname_path, input, training): 
#input: array containing the rasters of the workspace
    #debug =0
    if debug ==1:
        print("#### Start the Classification process ###")
        print(outname_path, input, training)
        
    base, ext =os.path.splitext(outname_path)
    gsg_path = base + ".gsg"
    temp_path = base + "raw"+ext
    if debug ==1:
        print(gsg_path, temp_path)
    arcpy.gp.CreateSignatures_sa(input,training,gsg_path,"COVARIANCE","typ")
    arcpy.gp.MLClassify_sa(input,gsg_path,temp_path,"0.0","EQUAL","#","#")
#here you can specify your reclassification to match your validation
    arcpy.gp.Reclassify_sa(temp_path,"Value","1 8;2 2;3 2;4 4;5 5;6 6;7 7;8 8;9 9",outname_path,"DATA")
    if debug==1:
        print("created: ",outname_path)
        print("#### Classification finished ###")

def validate (classification_path, validation):
    #This procedure takes a classification and a validation shapefile and creates a error matrix
    #the calculated kappa value is returned by the procedure
    createsubfolderifnotthere(os.path.abspath(os.curdir),"temp2")
    tempfolder = os.path.join(os.path.abspath(os.curdir),"temp2")
    #tempfolder="in_memory"
    #debug = 1
    if debug ==1:
        print("#### Starting the Validation process ###")
    base, ext = os.path.splitext (classification_path)
    tbl_path= base+"tbl.dbf"
    if debug ==1:
        print("classification path: ",classification_path)
        print("base: ",base)
        print("tablepath: ", tbl_path)
    oldworkspace = arcpy.env.workspace

    #arcpy.Delete_management(tempfolder) #INFDEF DIRECTORY FULL ERROR!
    arcpy.env.workspace = tempfolder
    if debug: print("start zonal Histogram")
    arcpy.gp.ZonalHistogram_sa(validation,"typ",classification_path,tbl_path,"#")
    if debug:
        print("zonalhist finished")
    db = dbf.Dbf(tbl_path)
    matrix = numpy.zeros((49), dtype=numpy.int).reshape((7,7))
    counter = 0
    checklist=[2,4,5,6,7,8,9]#Abfolge der ersten Spalte, wenn alle vorkommen
    for rec in db:
        counter2 = 0
        for value in rec:
            if (counter2 == 0):#check auf Nicht-Vorkommen
                if (int(value) == checklist [counter]):
                    n=1
                else:
                    while int(value)!=checklist [counter]:
                       counter+=1
            else:
                matrix[counter, counter2-1] = value
            counter2 += 1
        counter += 1
    #print "Die Error Matrix: "
    if debug==1:
        print(matrix)

    #Kappa Berechnung in Python mit numpy:
    #Alle Elemente
    n = float(matrix.sum())

    #summe xii
    rightpixels =0
    for zeile in range (7):
        rightpixels+=matrix[zeile,zeile]
    #print "alle richtigen Elemente: ", rightpixels

    #xi+ * x+i
    #b=0.
    #zeilensum = matrix.sum(axis=0)
    #spaltensum = matrix.sum(axis=1)
    #print zeilensum
    #print spaltensum
    #for zeile in range (7):
        #print "zeile: ",zeile
        #print (float(zeilensum[zeile])*float(spaltensum[zeile]))
        #b= b + (float(zeilensum[zeile])*float(spaltensum[zeile]))
        #print b
    #print "B: ",b

    ###Kappvergleich
    #Kappa aus n xii und xi+ * i+x
    khat = rightpixels / n
    #khat = ((n * rightpixels) - b)/((n * n) - b)
    if debug==1:
        print("Calculated Agreement Value: ",khat)
        print("#### Validation finished ###")


    arcpy.env.workspace = oldworkspace
    return khat


def upscale (alleraster, currentkombo, training, kappa):
    #debug =1
    createsubfolderifnotthere(os.path.abspath(os.curdir),"temp")
    tempfolder = os.path.join(os.path.abspath(os.curdir),"temp")
    kappalist=[]
    for raster in alleraster:
        if debug==1: print("Jetzt wird ", raster, " hinzugefuegt, klassifiziert und validiert")
        currentkombo.insert(0,raster)

        classify (os.path.join (tempfolder,"tempcls.tif"),currentkombo,training)
        kappacl = validate (os.path.join (tempfolder,"tempcls.tif"),validation)
        print(raster, "wurde hinzugefuegt, Kappa:", kappacl)
        kappalist.append (kappacl)
        currentkombo.remove(raster)
    if debug ==1:
        print(kappalist)
    if kappalist and ((max (kappalist)) >= kappa):
        print("Agreement verbessert von: ",kappa," auf: ", (max (kappalist)))
        if debug==1:
            print("kappalist:",kappalist)
            print("alte currentkombo:",currentkombo)
        #kappa wurde besser
        currentkombo.insert(0,alleraster [kappalist.index(max (kappalist))])
        if debug==1:
            print("neue currentkombo:", currentkombo)
        del alleraster [kappalist.index(max (kappalist))]
        kappa = upscale (alleraster, currentkombo, training, max(kappalist))
        return kappa
    else:
        print("FINISHED")
        print("Maximum Agreement reached: ",kappa)
        print("mit folgenden Rastern: ", currentkombo)
        createsubfolderifnotthere(inputfolder,"cl_agree")
        classify (inputfolder+"/cl_agree/upscale.tif",currentkombo, training)
        kappa = validate (inputfolder+"/cl_agree/upscale.tif", validation)
        textdatei = open (inputfolder+"/cl_agree/upscaleagree.txt", "w")
        textdatei.write("Results of the combiner:")
        textdatei.write("\n Klassifikation:")
        textdatei.write(inputfolder+"/cl_agree/upscale.tif")
        textdatei.write("\n Error Matrix:")
        textdatei.write(inputfolder+"/cl_agree/upscaletbl.dbf")
        textdatei.write("\n Agreemeent:")
        textdatei.write(str(kappa))
        textdatei.write("\n Diese Raster wurden verwendet:")
        for raster in currentkombo:
            textdatei.write(";\n")
            textdatei.write(raster)
        textdatei.write("\n folgende Raster ausgeschlossen:\n")
        for raster in alleraster:
            textdatei.write(raster)
            textdatei.write(";\n")
        textdatei.close()
        return kappa


#################Start Main Program###################
starttime = time.time()
arcpy.CheckOutExtension ("Spatial")
arcpy.env.overwriteOutput=True
arcpy.Delete_management ('in_memory')
#debuggiung on or not - as a global varaible
debug =0




#################Add your data here#####################
#Put your Training and Validation files here
training = "grounddata/training.shp"
validation = "grounddata/validation.shp"




inputfolderlist=[]

def listdir_fullpath(d):
    return [os.path.join(d, f) for f in os.listdir(d)]

#################Add your data here#####################
#Put the list of folders that contain the rasters to be classified
inputfolderlist=listdir_fullpath("c:/add/your/datahere")


print(inputfolderlist)

for inputfolder in inputfolderlist:
    arcpy.env.workspace = inputfolder
    rasters = arcpy.ListRasters()#All Rasters in the inputfolder
    currentlist =[]
    kappa = upscale (rasters, currentlist, training,0)
    #print kappa

    ###in globale Datei schreiben kappa, name, verwendete Raster
    textdatei = open (os.path.join(os.path.abspath(os.curdir),"global.txt"), "a")
    textdatei.write(str(kappa))
    textdatei.write(";")
    textdatei.write(str(inputfolder))
    textdatei.write(";\n")
    textdatei.close()

#Laufzeit bestimmen
endtime = time.time()
laufzeit = endtime-starttime
print(laufzeit, " seconds")
#tempfolder = os.path.join(os.path.abspath(os.curdir),"temp2")
#arcpy.Delete_management (tempfolder)
#tempfolder = os.path.join(os.path.abspath(os.curdir),"temp1")
#arcpy.Delete_management (tempfolder)
